﻿using Plugin.BLE.Abstractions.Contracts;
using Plugin.BLE.Abstractions.Utils;
using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Xamarin.Essentials;

namespace scgms_mobile.Devices.Generic
{
    /// <summary>
    ///   Base for device adapter
    /// </summary>
    public abstract class DeviceAdapterBase
    {
        // IDevice instance reference
        protected IDevice BLEDevice;
        // registered listener for updates
        protected DeviceValueListener Listener;

        /// <summary>
        ///   Constructor retaining device instance
        /// </summary>
        /// <param name="dev">Device instance</param>
        public DeviceAdapterBase(IDevice dev)
        {
            BLEDevice = dev;
            _BLEActionThread = new Thread(BLEActionThreadFunc);
            _BLEActionThread.Start();
        }

        /// <summary>
        ///   Registers a listener, unregistering the old one, if any
        /// </summary>
        /// <param name="listener">listener to be registered</param>
        public void RegisterListener(DeviceValueListener listener)
        {
            if (Listener != null)
                Listener.OnListenerUnregistered(this);

            Listener = listener;

            Listener.OnListenerRegistered(this);
        }

        public async Task<Boolean> UpdateDevice(IDevice dev)
        {
            BLEDevice = dev;
            return await ReInit();
        }

        /// <summary>
        /// Retrieves device unique ID (wrapped in Guid) if available
        /// </summary>
        /// <returns>Valid device Guid or zero Guid (Guid.Empty)</returns>
        public Guid GetDeviceGuid()
        {
            if (BLEDevice != null)
                return BLEDevice.Id;

            return Guid.Empty;
        }

        /// <summary>
        ///   Init method, this method is called after creating
        /// </summary>
#pragma warning disable CS1998 // Async method lacks 'await' operators and will run synchronously
        public virtual async Task<Boolean> Init() { /* stub */ return false; }
#pragma warning restore CS1998

        /// <summary>
        ///   Deinit method, this method is called before disposing
        /// </summary>
#pragma warning disable CS1998 // Async method lacks 'await' operators and will run synchronously
        public virtual async Task<Boolean> DeInit() { /* stub */ return true; }
#pragma warning restore CS1998

#pragma warning disable CS1998 // Async method lacks 'await' operators and will run synchronously
        public virtual async Task<Boolean> Forget() { /* stub */ return true; }
#pragma warning restore CS1998

        /// <summary>
        ///   Reinit method, this method is called after a new device connection has been established
        /// </summary>
#pragma warning disable CS1998 // Async method lacks 'await' operators and will run synchronously
        public virtual async Task<Boolean> ReInit() { /* stub */ return true; }
#pragma warning restore CS1998

#pragma warning disable CS1998 // Async method lacks 'await' operators and will run synchronously
        public virtual async Task<Boolean> Calibrate(double valueMmolL) { /* stub */ return true; }
#pragma warning restore CS1998

#pragma warning disable CS1998 // Async method lacks 'await' operators and will run synchronously
        public virtual async Task<Boolean> SetupSensor(bool isNew) { /* stub */ return true; }
#pragma warning restore CS1998

#pragma warning disable CS1998 // Async method lacks 'await' operators and will run synchronously
        public virtual async Task<Boolean> ReportLastSeqNumber(uint seq) { /* stub */ return true; }
#pragma warning restore CS1998

        private enum BLEActionType
        {
            Read,
            Write,
        }

        private class BLEActionItem
        {
            public BLEActionType Type { get; set; }

            public ICharacteristic Characteristic { get; set; }

            public byte[] Data { get; set; }

            public object Awaitable { get; set; }

            public bool Success { get; set; } = false;
        }

        private List<BLEActionItem> _BLEQueue = new List<BLEActionItem>();
        private Thread _BLEActionThread;
        private object BLEActionLocker = new object();

        private void BLEActionThreadFunc()
        {
            while (true)
            {
                BLEActionItem task;
                lock (BLEActionLocker)
                {
                    while (_BLEQueue.Count == 0)
                        Monitor.Wait(BLEActionLocker);
                    task = _BLEQueue[0];
                    _BLEQueue.RemoveAt(0);
                }

                if (task == null)
                    return;

                if (task.Type == BLEActionType.Read)
                {
                    try
                    {
                        var taskentry = task.Characteristic.ReadAsync();
                        taskentry.Wait();

                        task.Data = taskentry.Result;
                        task.Success = true;
                    }
                    catch
                    {
                        task.Success = false;
                    }

                    if (task.Awaitable != null)
                    {
                        lock (task.Awaitable)
                        {
                            Monitor.PulseAll(task.Awaitable);
                        }
                    }
                }
                else if (task.Type == BLEActionType.Write)
                {
                    try
                    {
                        var taskentry = MainThread.InvokeOnMainThreadAsync(() =>
                        {
                            return task.Characteristic.WriteAsync(task.Data);
                        });
                        taskentry.Wait();
                        task.Success = true;
                    }
                    catch
                    {
                        task.Success = false;
                    }

                    if (task.Awaitable != null)
                    {
                        lock (task.Awaitable)
                        {
                            Monitor.PulseAll(task.Awaitable);
                        }
                    }
                }
            }
        }

        private void EnqueueBLETask(BLEActionItem task, bool prioritize = false)
        {
            lock (BLEActionLocker)
            {
                if (prioritize)
                    _BLEQueue.Insert(0, task);
                else
                    _BLEQueue.Add(task);
                Monitor.PulseAll(BLEActionLocker);
            }
        }

        public class ReadResult
        {
            public bool Success { get; set; }
            public byte[] Data { get; set; }
        }

        protected Task<ReadResult> QueuedRead(ICharacteristic chr, bool prioritize = false)
        {
            var tsk = Task.Run(() => {

                ReadResult rr = new ReadResult();

                object awaitable = new object();
                lock (awaitable)
                {
                    var item = new BLEActionItem
                    {
                        Type = BLEActionType.Read,
                        Characteristic = chr,
                        Awaitable = awaitable
                    };

                    EnqueueBLETask(item, prioritize);

                    if (Monitor.Wait(awaitable, 3000))
                    {
                        rr.Success = item.Success;
                        rr.Data = item.Data;
                        return rr;
                    }
                }

                return null;
            });

            return tsk;
        }

        protected Task QueuedWrite(ICharacteristic chr, byte[] data, bool prioritize = false)
        {
            var tsk = Task.Run(() => {

                object awaitable = new object();
                lock (awaitable)
                {
                    var item = new BLEActionItem
                    {
                        Type = BLEActionType.Write,
                        Characteristic = chr,
                        Data = data,
                        Awaitable = awaitable
                    };

                    EnqueueBLETask(item, prioritize);

                    Monitor.Wait(awaitable, 3000);
                }
            });

            return tsk;
        }

    }
}
