﻿using Microsoft.AppCenter.Crashes;
using Plugin.BLE.Abstractions.Contracts;
using Plugin.BLE.Abstractions.EventArgs;
using scgms_mobile.Misc;
using scgms_mobile.SCGMSService.Impl.Messages;
using System;
using System.Diagnostics;
using System.Security.Cryptography;
using System.Threading.Tasks;
using Xamarin.Essentials;
using Xamarin.Forms;

namespace scgms_mobile.Devices.Guardian
{
    /// <summary>
    ///   Implementation of Medtronic Guardian Connect device
    /// </summary>
    class GuardianDevice : Generic.DeviceAdapterBase
    {
        private IService CGMService;
        private ICharacteristic MeasurementReadyChr;
        private ICharacteristic CalResponseReadyChr;
        private ICharacteristic BackfillReadyChr;
        private ICharacteristic MeasurementDataChr;
        private ICharacteristic CalResponseDataChr;
        private ICharacteristic BackfillDataChr;
        private ICharacteristic NonCritOpsChr;
        private ICharacteristic CritOpsChr;
        private ICharacteristic SensorConnStateChr;
        private ICharacteristic StartCommsChr;
        private ICharacteristic SensorCalExpirationChr;
        private ICharacteristic SensorExpirationChr;

        private IService BatteryService;
        private ICharacteristic BatteryLevelChr;

        private Support.Aes128CounterMode CryptoApp;
        private Support.Aes128CounterMode CryptoSensor;

        private ICryptoTransform EncryptorApp, EncryptorSensor;
        private ICryptoTransform DecryptorApp, DecryptorSensor;

        private System.Threading.Timer PeriodicUpdater;

        //private readonly string CertIssuer = "https://carelink.minimed.eu";
        //private readonly string CertIssuer2 = "https://carelink.minimed.com";
        private readonly byte[] PresharedSensorKey = new byte[] { 0x71, 0xBF, 0xAA, 0x24, 0x38, 0x6E, 0x57, 0x65, 0x6F, 0x0B, 0x28, 0x5E, 0x22, 0x3F, 0x0F, 0xB6 };
        private readonly byte[] PresharedGuardianKey = new byte[] { 0xAF, 0xBC, 0xF4, 0xAF, 0x44, 0x59, 0xF8, 0x70, 0x7A, 0x17, 0xDA, 0xD2, 0x42, 0x12, 0x32, 0xFF };

        private byte[] sensorAdvKey = new byte[16];

        private byte[] lastSentCritOpsPacketBytes;
        private byte[] lastSentNonCritOpsPacketBytes;

        private byte[] measDataIVSeed;

        private bool sensorPaired = false;
        private bool timeSynced = false;
        private bool forceTimeSync = false;

        private double SensorTimeStart = -1.0;

        private uint LastReportedSeq = 0;
        private bool ForceBackfillRequest = false;
        private DateTime? BackfillPendingSince = null;

        private DateTime LastSendTime = DateTime.UtcNow;
        private DateTime SensorExpirationLastRefresh = DateTime.UtcNow - new TimeSpan(1,0,0); // subtract one hour to trigger refreshing
        private bool SensorExpiryValid = false;
        private bool SensorCalExpiryValid = false;

        private bool _BackfillInProgress = false;
        private bool BackfillInProgress { get { return _BackfillInProgress; }
            set
            {
                _BackfillInProgress = value;
                if (Listener != null)
                    Listener.SetBackfillProcessState(_BackfillInProgress);
            }
        }

        private class StoredMeasurement
        {
            public int status;
            public uint sequence;
            public ushort glucose;
            public long timestamp;
        }

        private StoredMeasurement DeferredMeasurement = null;

        /// <summary>
        /// IEEE Device constructor
        /// </summary>
        /// <param name="dev">device to be wrapped</param>
        public GuardianDevice(IDevice dev) : base(dev)
        {
            // AES128/CTR/NoPadding
            byte[] iv, commKey;
            byte[] ivSens;

            Random rnd = new Random();

            sensorPaired = true;

            iv = new byte[16];
            rnd.NextBytes(iv);
            if ((commKey = StorageMgr.GetAppCommKey()) == null)
            {
                commKey = new byte[16];
                Debug.WriteLine("AppCommKey not present, generating new key!");

                // see below
                //rnd.NextBytes(commKey);

                sensorPaired = false;
            }
            // well, this should be random, but for debugging purposes, let's leave it this way
            for (int idx = 0; idx < 16; idx++)
                commKey[idx] = (byte)idx;

            var tval = StorageMgr.GetSensorStartTime();
            if (tval.HasValue)
                SensorTimeStart = tval.Value;
            else
            {
                Debug.WriteLine("SensorTimeStart not present, forcing time sync");
                sensorPaired = false;
                timeSynced = false;
                forceTimeSync = true;
            }

            ivSens = new byte[16];
            rnd.NextBytes(ivSens);

            CryptoApp = new Support.Aes128CounterMode(iv, commKey);
            CryptoApp.Padding = System.Security.Cryptography.PaddingMode.None;

            StorageMgr.SetAppCommKey(CryptoApp.GetKey());

            CryptoSensor = new Support.Aes128CounterMode(ivSens, PresharedGuardianKey);
            CryptoSensor.Padding = System.Security.Cryptography.PaddingMode.None;
        }

        public override async Task<Boolean> ReInit()
        {
            await DeInit();
            return await Init();
        }

        public override async Task<Boolean> DeInit()
        {
            if (PeriodicUpdater != null)
            {
                PeriodicUpdater.Dispose();
                PeriodicUpdater = null;
            }

            StorageMgr.SetAppCommIV(((Support.CounterModeCryptoTransform)EncryptorApp).GetCounter());
            StorageMgr.SetSensorCommIV(((Support.CounterModeCryptoTransform)EncryptorSensor).GetCounter());

            try
            {
                // these will fail if the device has already been disconnected prior to DeInit call

                await MeasurementReadyChr.StopUpdatesAsync();
                await CalResponseReadyChr.StopUpdatesAsync();
                await NonCritOpsChr.StopUpdatesAsync();
                await CritOpsChr.StopUpdatesAsync();
                await SensorConnStateChr.StopUpdatesAsync();
                await BatteryLevelChr.StopUpdatesAsync();
                await StartCommsChr.StopUpdatesAsync();
                await SensorCalExpirationChr.StopUpdatesAsync();
                await SensorExpirationChr.StopUpdatesAsync();
            }
            catch (Exception)
            {
                //
            }

            return true;
        }

        public override async Task<Boolean> Forget()
        {
            // forgetting sensor means generating whole new key on next attempt to handshake
            StorageMgr.SetAppCommKey(null);

            await DeInit();

            return true;
        }

        /// <summary>
        /// Initializes adapter, subscribes for updates, etc.
        /// </summary>
        public override async Task<Boolean> Init()
        {
            CGMService = await BLEDevice.GetServiceAsync(BLEKnownGuids.GuardianMainServiceGUID);
            MeasurementReadyChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_MeasurementReady);
            CalResponseReadyChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_CalibrationResponseReady);
            BackfillReadyChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_BackfillReady);
            StartCommsChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_START_COMMS);
            SensorConnStateChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_SENSOR_CONNECTED_STATE);
            SensorCalExpirationChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_SENSOR_CAL_EXPIRATION);
            SensorExpirationChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_SENSOR_EXPIRATION);
            MeasurementDataChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_MeasurementData);
            CalResponseDataChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_CalibrationResponseData);
            BackfillDataChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_BackfillData);
            NonCritOpsChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_NonCriticalOpsControlPoint);
            CritOpsChr = await CGMService.GetCharacteristicAsync(BLEKnownGuids.Guardian_CriticalOpsControlPoint);

            BatteryService = await BLEDevice.GetServiceAsync(BLEKnownGuids.BatteryServiceGUID);
            BatteryLevelChr = await BatteryService.GetCharacteristicAsync(BLEKnownGuids.BatteryChargeGUID);

            // subscribe for measured value changes
            MeasurementReadyChr.ValueUpdated += OnMeasuredValueIndicated;
            await MeasurementReadyChr.StartUpdatesAsync();

            CalResponseReadyChr.ValueUpdated += OnCalResponseValueIndicated;
            await CalResponseReadyChr.StartUpdatesAsync();

            BackfillReadyChr.ValueUpdated += OnBackfillReadyValueIndicated;
            await BackfillReadyChr.StartUpdatesAsync();

            StartCommsChr.ValueUpdated += OnStartCommsValueIndicated;
            await StartCommsChr.StartUpdatesAsync();

            NonCritOpsChr.ValueUpdated += OnNonCritOpsIndicated;
            await NonCritOpsChr.StartUpdatesAsync();

            CritOpsChr.ValueUpdated += OnCritOpsIndicated;
            await CritOpsChr.StartUpdatesAsync();

            SensorConnStateChr.ValueUpdated += OnSensorConnStateValueIndicated;
            await SensorConnStateChr.StartUpdatesAsync();

            SensorCalExpirationChr.ValueUpdated += OnSensorCalExpirationValueIndicated;
            await SensorCalExpirationChr.StartUpdatesAsync();

            SensorExpirationChr.ValueUpdated += OnSensorExpirationValueIndicated;
            await SensorExpirationChr.StartUpdatesAsync();

            BatteryLevelChr.ValueUpdated += OnBatteryLevelValueIndicated;
            await BatteryLevelChr.StartUpdatesAsync();

            EncryptorApp = CryptoApp.CreateEncryptor(CryptoApp.GetKey(), null);
            DecryptorApp = CryptoApp.CreateDecryptor(CryptoApp.GetKey(), null);

            EncryptorSensor = CryptoSensor.CreateEncryptor(PresharedGuardianKey, null);
            DecryptorSensor = CryptoSensor.CreateDecryptor(PresharedGuardianKey, null);

            // always trigger time resync
            timeSynced = false;

            if (Listener != null)
                Listener.OnAdapterStatusChange(Generic.DeviceAdapterStatus.InitComm);

            if (Listener != null)
                Listener.OnStatusChange(AppSensorStatus.Initializing);

            await Send_NonCritOps(Packets.StartComms.Build(), false);

            return true;
        }

        private uint UpdateTick = 0;

        private async void PeriodicUpdate(object ob)
        {
            int scheduleNextIn = 10000;

            try
            {

                if (!timeSynced)
                {
                    SyncTime();
                }
                else if (ForceBackfillRequest || BackfillPendingSince.HasValue)
                {
                    if (!BackfillInProgress)
                    {
                        if (Listener != null)
                            Listener.OnAdapterStatusChange(Generic.DeviceAdapterStatus.BackfillPhase);

                        BackfillPendingSince = DateTime.Now;
                        ForceBackfillRequest = false;
                        await Send_NonCritOps(Packets.RequestBackfillRange.Build(LastReportedSeq, 0xFFFFFFFF), true);
                    }
                }
                else
                {
                    bool recentComm = (DateTime.UtcNow.Subtract(LastSendTime).TotalSeconds < 10);

                    if (UpdateTick % 3 == 0)
                    {
                        if (!recentComm)
                        {
                            await ReadCharacteristic(SensorConnStateChr);
                            scheduleNextIn = 2000;
                        }
                    }
                    else if (UpdateTick % 3 == 1)
                    {
                        if (!recentComm)
                        {
                            await ReadCharacteristic(BatteryLevelChr);
                            scheduleNextIn = 2000;
                        }
                    }
                    else
                    {
                        try
                        {
                            if (measDataIVSeed != null)
                                await ReadMeasurement();
                            else
                            {
                                /*if (!recentComm)
                                    await ReadCharacteristic(MeasurementDataChr);*/
                            }
                        }
                        catch (Exception ex)
                        {
                            Debug.WriteLine("Could not read measurement/ready flag, exception: " + ex.Message);
                        }

                        // if the expiry dates are valid, schedule next iteration in 60 seconds
                        if (SensorCalExpiryValid && SensorExpiryValid)
                            scheduleNextIn = 60000;
                    }

                    UpdateTick++;
                }
            }
            catch (Exception ex)
            {
                // oh, don't judge me!
                Debug.WriteLine("Periodic updater exception: " + ex.Message);
                Crashes.TrackError(ex);
            }

            if (PeriodicUpdater != null)
            {
                PeriodicUpdater.Dispose();
                PeriodicUpdater = null;
            }
            PeriodicUpdater = new System.Threading.Timer(new System.Threading.TimerCallback(PeriodicUpdate), null, scheduleNextIn, System.Threading.Timeout.Infinite);
        }

        public override async Task<Boolean> Calibrate(double valueMmolL)
        {
            uint timeOffset = (uint)(DateTime.UtcNow.Subtract(new DateTime(1970, 1, 1)).TotalSeconds - SensorTimeStart);

            await Send_CritOps(Packets.CalibrationRequest.Build(valueMmolL, timeOffset), true);

            return true;
        }

        public override async Task<Boolean> SetupSensor(bool isNew)
        {
            SensorType type = isNew ? SensorType.NEW : SensorType.EXISTING;

            await Send_CritOps(Packets.SetSensorType.Build(type), true);

            return true;
        }

#pragma warning disable CS1998 // Async method lacks 'await' operators and will run synchronously
        public override async Task<Boolean> ReportLastSeqNumber(uint seq)
        {
            // TODO: determine current seq somehow

            LastReportedSeq = seq;
            ForceBackfillRequest = true;

            return true;
        }
#pragma warning restore CS1998 // Async method lacks 'await' operators and will run synchronously

        /// <summary>
        /// Parses input from CGM Measurement characteristics and updates listener
        /// </summary>
        /// <param name="input">received bytes from CGM measurement characteristics</param>
        private void UpdateMeasuredValue(double measurement, long timestamp, uint seq, bool valid)
        {
            measurement /= 18.018; // mg/dl --> mmol/l

            // update listener
            if (Listener != null)
                Listener.OnNewMeasurement(Generic.DeviceSignalType.Interstitial_Glucose, measurement, timestamp, seq, valid);
        }

        private bool ProcessMeasurement(int sensorStatus, uint seq, ushort glucose, long timestamp)
        {
            try
            {
                if ((sensorStatus & (int)SensorStatus.TIME_SYNC_NEEDED) != 0)
                    timeSynced = false;

                if ((sensorStatus & (int)SensorStatus.CHANGE_SENSOR) != 0)
                {
                    if (Listener != null)
                        Listener.OnStatusChange(AppSensorStatus.NeedsChangeSensor);
                }
                else if ((sensorStatus & (int)SensorStatus.SENSOR_CONFIGURATION_NEEDED) != 0)
                {
                    if (Listener != null)
                        Listener.OnStatusChange(AppSensorStatus.NeedsSetup);
                }
                else if ((sensorStatus & (int)SensorStatus.SENSOR_IN_INITIALIZATION) != 0)
                {
                    if (Listener != null)
                        Listener.OnStatusChange(AppSensorStatus.Warmup);
                }
                else if ((sensorStatus & (int)SensorStatus.CALIBRATION_REQUIRED) != 0)
                {
                    if (Listener != null)
                        Listener.OnStatusChange(AppSensorStatus.NeedsCalibration);
                }
                else if ((sensorStatus & (int)SensorStatus.CALIBRATION_PENDING) != 0)
                {
                    if (Listener != null)
                        Listener.OnStatusChange(AppSensorStatus.CalibrationPending);
                }
                // TODO: handle the rest of error statuses correctly
                else
                {
                    if (Listener != null)
                        Listener.OnStatusChange(AppSensorStatus.Connected);
                }

                // 0xFFFF is special value marking "no value available, but still the flow works as expected" - in this case, only status flags are relevant for us
                UpdateMeasuredValue((double)glucose, timestamp, seq, (glucose < 0x0200));
            }
            catch (Exception ex)
            {
                // oh, don't judge me (2)!
                Debug.WriteLine("ProcessMeasurement exception: " + ex.Message);
                Crashes.TrackError(ex);
            }

            return true;
        }

        private async Task<UInt16> ReadMeasurement()
        {
            try
            {
                ((Support.CounterModeCryptoTransform)DecryptorApp).SetCounter(measDataIVSeed);

                byte[] measurement = await ReadCharacteristic(MeasurementDataChr);
                if (measurement != null && measurement.Length > 10)
                {
                    Debug.WriteLine("< MEASUREMENT_DATA read: " + Utils.ByteArrayToHexString(measurement));
                    MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "MEASUREMENT_DATA", DataString = Utils.ByteArrayToHexString(measurement) }, SensorCommRecordMessage.Name);

                    ChrValueWrapper chrval2 = new ChrValueWrapper(measurement);

                    int flags = (int)chrval2.ReadUShort();

                    bool ivseed_flag = GuardianUtils.GetMaskedBits(32768, 15, flags) == 1;
                    bool unk_flag = GuardianUtils.GetMaskedBits(16384, 14, flags) == 1;
                    int authLevel = GuardianUtils.GetMaskedBits(15872, 9, flags);
                    int unk_val = GuardianUtils.GetMaskedBits(511, 0, flags);

                    uint seq = chrval2.ReadUInt();
                    uint timestamp_offset = chrval2.ReadUInt();

                    long timestamp = ((long)SensorTimeStart + (long)timestamp_offset);

                    Debug.WriteLine("ivseed_flag = " + ivseed_flag + ", unk_flag = " + unk_flag + ", authLevel = " + authLevel + ", unk_val = " + unk_val);
                    // we don't really need those flags here, do we?

                    byte[] outbuffer = new byte[measurement.Length - 10];
                    DecryptorApp.TransformBlock(measurement, 10, measurement.Length - 10, outbuffer, 0);

                    measDataIVSeed = null;

                    Debug.WriteLine("< MEASUREMENT_DATA decrypted: " + Utils.ByteArrayToHexString(outbuffer));
                    MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "MEASUREMENT_DATA(decrypted)", DataString = Utils.ByteArrayToHexString(outbuffer) }, SensorCommRecordMessage.Name);

                    ChrValueWrapper decrval = new ChrValueWrapper(outbuffer);

                    byte unk_meas_flag1 = decrval.ReadByte();
                    byte signal_type1 = decrval.ReadByte();
                    byte signal_type2 = decrval.ReadByte();
                    ushort glucose = decrval.ReadUShort();
                    int sensorStatus = (int)decrval.ReadUShort();
                    ushort rateOfChange = decrval.ReadUShort();

                    Debug.WriteLine("unk_flag = " + unk_meas_flag1 + ", signal_type1 = " + signal_type1 + ", signal_type2 = " + signal_type2 + ", glucose = " + glucose + ", sensorStatus = " + sensorStatus + ", rateOfChange = " + rateOfChange);

                    if (BackfillPendingSince != null || LastReportedSeq + 1 < seq)
                    {
                        DeferredMeasurement = new StoredMeasurement();
                        DeferredMeasurement.status = sensorStatus;
                        DeferredMeasurement.sequence = seq;
                        DeferredMeasurement.glucose = glucose;
                        DeferredMeasurement.timestamp = timestamp;

                        Debug.WriteLine("Value storing deferred due to backfill in progress");

                        // force backfill update on seq mismatch
                        if (BackfillPendingSince == null)
                        {
                            ForceBackfillRequest = true;
                            PeriodicUpdate(null);
                        }
                    }
                    else
                    {
                        // do not send already sent sequence numbers
                        if (LastReportedSeq != seq)
                        {
                            ProcessMeasurement(sensorStatus, seq, glucose, timestamp);
                            LastReportedSeq = seq;
                        }
                        else
                            Debug.WriteLine("Sequence " + seq + " discarded - such value was already received");
                    }

                    return glucose;
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Could not read and store value, exception: " + ex.Message);
                Crashes.TrackError(ex);
            }

            return 0xFFFF;
        }

        /// <summary>
        /// Callback from BLE module - called upon receiving a new value from measurement characteristics
        /// </summary>
        /// <param name="o">source object</param>
        /// <param name="args">event arguments</param>
        private async void OnMeasuredValueIndicated(object o, CharacteristicUpdatedEventArgs args)
        {
            try
            {
                var value = args.Characteristic.Value;

                Debug.WriteLine("< MEASUREMENT_READY data: " + Utils.ByteArrayToHexString(value));
                MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "MEASUREMENT_READY", DataString = Utils.ByteArrayToHexString(value) }, SensorCommRecordMessage.Name);

                System.Threading.Thread.Sleep(1000);

                ChrValueWrapper chrval = new ChrValueWrapper(value);

                int flags = (int)chrval.ReadUShort();

                bool ivseed_flag = GuardianUtils.GetMaskedBits(32768, 15, flags) == 1;
                bool unk_flag = GuardianUtils.GetMaskedBits(16384, 14, flags) == 1;
                int authLevel = GuardianUtils.GetMaskedBits(15872, 9, flags);
                int unk_val = GuardianUtils.GetMaskedBits(511, 0, flags);

                Debug.WriteLine("ivseed_flag = " + ivseed_flag + ", unk_flag = " + unk_flag + ", authLevel = " + authLevel + ", unk_val = " + unk_val);

                if (ivseed_flag)
                {
                    measDataIVSeed = new byte[16];
                    Array.Copy(value, 4, measDataIVSeed, 0, 16);
                }

                await ReadMeasurement();
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Exception occurred (1): " + ex.Message);
                Crashes.TrackError(ex);
            }
        }

        private async void OnCalResponseValueIndicated(object o, CharacteristicUpdatedEventArgs args)
        {
            try
            {
                var value = args.Characteristic.Value;

                Debug.WriteLine("< CAL_RESPONSE_READY data: " + Utils.ByteArrayToHexString(value));
                MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "CAL_RESPONSE_READY", DataString = Utils.ByteArrayToHexString(value) }, SensorCommRecordMessage.Name);

                ChrValueWrapper chrval = new ChrValueWrapper(value);

                int flags = (int)chrval.ReadUShort();

                bool ivseed_flag = GuardianUtils.GetMaskedBits(32768, 15, flags) == 1;
                bool unk_flag = GuardianUtils.GetMaskedBits(16384, 14, flags) == 1;
                int authLevel = GuardianUtils.GetMaskedBits(15872, 9, flags);
                int unk_val = GuardianUtils.GetMaskedBits(511, 0, flags);

                Debug.WriteLine("ivseed_flag = " + ivseed_flag + ", unk_flag = " + unk_flag + ", authLevel = " + authLevel + ", unk_val = " + unk_val);

                if (ivseed_flag)
                {
                    byte[] iv = new byte[16];
                    Array.Copy(value, 4, iv, 0, 16);

                    ((Support.CounterModeCryptoTransform)DecryptorApp).SetCounter(iv);
                }

                try
                {
                    byte[] measurement = await ReadCharacteristic(CalResponseDataChr);
                    if (measurement != null && measurement.Length > 10)
                    {
                        Debug.WriteLine("< CAL_RESPONSE_DATA read: " + Utils.ByteArrayToHexString(measurement));
                        MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "CAL_RESPONSE_DATA", DataString = Utils.ByteArrayToHexString(measurement) }, SensorCommRecordMessage.Name);

                        ChrValueWrapper chrval2 = new ChrValueWrapper(measurement);

                        flags = (int)chrval2.ReadUShort();

                        ivseed_flag = GuardianUtils.GetMaskedBits(32768, 15, flags) == 1;
                        unk_flag = GuardianUtils.GetMaskedBits(16384, 14, flags) == 1;
                        authLevel = GuardianUtils.GetMaskedBits(15872, 9, flags);
                        unk_val = GuardianUtils.GetMaskedBits(511, 0, flags);

                        Debug.WriteLine("ivseed_flag = " + ivseed_flag + ", unk_flag = " + unk_flag + ", authLevel = " + authLevel + ", unk_val = " + unk_val);

                        // TODO: extract bits from header

                        // TODO: fix this

                        byte[] outbuffer = new byte[measurement.Length - 10];
                        DecryptorApp.TransformBlock(measurement, 10, measurement.Length - 10, outbuffer, 0);

                        Debug.WriteLine("< CAL_RESPONSE_DATA decrypted: " + Utils.ByteArrayToHexString(outbuffer));
                        MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "CAL_RESPONSE_DATA(decrypted)", DataString = Utils.ByteArrayToHexString(outbuffer) }, SensorCommRecordMessage.Name);

                        // 04 01 8800 03
                        ChrValueWrapper decrval = new ChrValueWrapper(outbuffer);
                        decrval.ReadByte();     // unk, discard
                        decrval.ReadByte();     // unk, discard
                        decrval.ReadUShort();   // TODO: use this glucose level
                        CalibrationStatus result = (CalibrationStatus)decrval.ReadByte();

                        switch (result)
                        {
                            case CalibrationStatus.CALIBRATION_ACCEPTED: if (Listener != null) Listener.OnCalibrationResult(Generic.Miscellanous.SensorCalibrationResult.Accepted); break;
                            case CalibrationStatus.CALIBRATION_REJECTED: if (Listener != null) Listener.OnCalibrationResult(Generic.Miscellanous.SensorCalibrationResult.Rejected); break;
                            case CalibrationStatus.CALIBRATION_NOT_CALCULATED: if (Listener != null) Listener.OnCalibrationResult(Generic.Miscellanous.SensorCalibrationResult.AcceptedNotCalculated); break;
                        }
                    }
                }
                catch (Exception ex)
                {
                    Debug.WriteLine("Could not read calibration response, exception: " + ex.Message);
                    Crashes.TrackError(ex);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Exception occurred (2): " + ex.Message);
                Crashes.TrackError(ex);
            }
        }

        private async void OnBackfillReadyValueIndicated(object o, CharacteristicUpdatedEventArgs args)
        {
            try
            {
                var value = args.Characteristic.Value;

                Debug.WriteLine("< BACKFILL_READY data: " + Utils.ByteArrayToHexString(value));
                MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "BACKFILL_READY", DataString = Utils.ByteArrayToHexString(value) }, SensorCommRecordMessage.Name);

                BackfillInProgress = true;

                ChrValueWrapper chrval = new ChrValueWrapper(value);

                int flags = (int)chrval.ReadUShort();

                bool ivseed_flag = GuardianUtils.GetMaskedBits(32768, 15, flags) == 1;
                bool unk_flag = GuardianUtils.GetMaskedBits(16384, 14, flags) == 1;
                int authLevel = GuardianUtils.GetMaskedBits(15872, 9, flags);
                int unk_val = GuardianUtils.GetMaskedBits(511, 0, flags);

                Debug.WriteLine("ivseed_flag = " + ivseed_flag + ", unk_flag = " + unk_flag + ", authLevel = " + authLevel + ", unk_val = " + unk_val);

                //if (ivseed_flag)
                {
                    byte[] iv = new byte[16];
                    Array.Copy(value, 4, iv, 0, 16);

                    ((Support.CounterModeCryptoTransform)DecryptorApp).SetCounter(iv);
                }

                try
                {
                    byte[] measurement = await ReadCharacteristic(BackfillDataChr);
                    if (measurement != null)
                    {
                        Debug.WriteLine("< BACKFILL_DATA data: " + Utils.ByteArrayToHexString(measurement));
                        MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "BACKFILL_DATA", DataString = Utils.ByteArrayToHexString(measurement) }, SensorCommRecordMessage.Name);

                        ChrValueWrapper dataval = new ChrValueWrapper(measurement);

                        flags = (int)dataval.ReadUShort();

                        uint seq = dataval.ReadUInt();
                        uint measTime = dataval.ReadUInt();
                        Debug.WriteLine("SEQ = " + seq + ", measTime = " + measTime);

                        // end backfill range marker
                        if (seq == 0xFFFFFFFF || measTime == 0xFFFFFFFF || measurement.Length <= 10)
                        {
                            BackfillPendingSince = null;

                            if (Listener != null)
                                Listener.OnAdapterStatusChange(Generic.DeviceAdapterStatus.Connected);

                            if (DeferredMeasurement != null && DeferredMeasurement.sequence != LastReportedSeq - 1)
                            {
                                BackfillInProgress = false;
                                ProcessMeasurement(DeferredMeasurement.status, DeferredMeasurement.sequence, DeferredMeasurement.glucose, DeferredMeasurement.timestamp);
                                LastReportedSeq = DeferredMeasurement.sequence;
                                DeferredMeasurement = null;
                            }
                            else
                                LastReportedSeq = LastReportedSeq - 1;

                            BackfillInProgress = false;
                            return;
                        }

                        byte[] outbuffer = new byte[measurement.Length - 10];
                        DecryptorApp.TransformBlock(measurement, 10, measurement.Length - 10, outbuffer, 0);

                        Debug.WriteLine("< BACKFILL_DATA decrypted: " + Utils.ByteArrayToHexString(outbuffer));
                        MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "BACKFILL_DATA(decrypted)", DataString = Utils.ByteArrayToHexString(outbuffer) }, SensorCommRecordMessage.Name);

                        ChrValueWrapper decval = new ChrValueWrapper(outbuffer);

                        int recSize = decval.ReadByte();
                        uint seqItr = seq;

                        for (int i = 0; i < outbuffer.Length - 1; i += recSize)
                        {
                            ushort timeOffset = decval.ReadUShort();
                            byte unk1 = decval.ReadByte();
                            byte signalType1 = decval.ReadByte();
                            byte signalType2 = decval.ReadByte();
                            ushort glucose = decval.ReadUShort();
                            int sensorStatus = (int)decval.ReadUShort();

                            long timestamp = ((long)SensorTimeStart + (long)measTime + (long)timeOffset);

                            Debug.WriteLine("Backfill, seq: " + seqItr + ", glucose: " + glucose);
                            ProcessMeasurement(sensorStatus, seqItr, glucose, timestamp);

                            seqItr++;
                        }

                        LastReportedSeq = seqItr;
                        ForceBackfillRequest = true;
                        BackfillInProgress = false;
                        PeriodicUpdate(null);
                    }
                }
                catch (Exception ex)
                {
                    BackfillInProgress = false;
                    Debug.WriteLine("Could not read and store backfill value, exception: " + ex.Message);
                    Crashes.TrackError(ex);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Exception occurred (3): " + ex.Message);
                Crashes.TrackError(ex);
            }

            BackfillInProgress = false;
        }

        private async void OnSensorConnStateValueIndicated(object o, CharacteristicUpdatedEventArgs args)
        {
            try
            {
                var value = args.Characteristic.Value;

                Debug.WriteLine("< SENSOR_CONNECTED_STATE data: " + Utils.ByteArrayToHexString(value));
                MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "SENSOR_CONNECTED_STATE", DataString = Utils.ByteArrayToHexString(value) }, SensorCommRecordMessage.Name);

                if (value.Length == 0)
                    return;

                byte connState = value[0];
                if (connState == 1)
                {
                    long timeRefresh = (SensorCalExpiryValid && SensorExpiryValid) ? 5 * 60 : 30;

                    if (DateTime.UtcNow.Subtract(SensorExpirationLastRefresh).TotalSeconds >= timeRefresh)
                    {
                        await Send_NonCritOps(Packets.RequestCalibrationExpiration.Build(), true);
                        await Send_NonCritOps(Packets.RequestSensorExpiration.Build(), true);
                    }
                }
                else
                {
                    if (Listener != null)
                        Listener.OnStatusChange(AppSensorStatus.Disconnected);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Exception occurred (4): " + ex.Message);
                Crashes.TrackError(ex);
            }
        }

        private void OnSensorExpirationValueIndicated(object o, CharacteristicUpdatedEventArgs args)
        {
            try
            {
                var value = args.Characteristic.Value;
                ChrValueWrapper chrvalue = new ChrValueWrapper(value);

                Debug.WriteLine("< SENSOR_EXPIRATION data: " + Utils.ByteArrayToHexString(value));
                MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "SENSOR_EXPIRATION", DataString = Utils.ByteArrayToHexString(value) }, SensorCommRecordMessage.Name);

                DateTime? sensorDatetimeStart = null;

                uint offset = chrvalue.ReadUInt();
                if (offset == 0xFFFFFFFF)
                {
                    SensorExpiryValid = false;
                    Debug.WriteLine("No expiration date/time received - sensor needs setup");

                    if (Listener != null)
                        Listener.OnStatusChange(AppSensorStatus.NeedsSetup);
                }
                else
                {
                    SensorExpiryValid = true;
                    sensorDatetimeStart = new DateTime(1970, 1, 1, 0, 0, 0, 0, System.DateTimeKind.Utc);
                    sensorDatetimeStart = sensorDatetimeStart.Value.AddSeconds(SensorTimeStart + offset).ToLocalTime();

                    Debug.WriteLine("Sensor will expire at date/time: " + sensorDatetimeStart.Value.ToLongDateString() + ", " + sensorDatetimeStart.Value.ToLongTimeString());
                }

                if (Listener != null)
                    Listener.OnSensorExpireDateChanged(sensorDatetimeStart);
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Exception occurred (5): " + ex.Message);
                Crashes.TrackError(ex);
            }
        }

        private void OnSensorCalExpirationValueIndicated(object o, CharacteristicUpdatedEventArgs args)
        {
            try
            {
                var value = args.Characteristic.Value;
                ChrValueWrapper chrvalue = new ChrValueWrapper(value);

                Debug.WriteLine("< SENSOR_CAL_EXPIRATION data: " + Utils.ByteArrayToHexString(value));
                MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "SENSOR_CAL_EXPIRATION", DataString = Utils.ByteArrayToHexString(value) }, SensorCommRecordMessage.Name);

                DateTime? sensorCalExpire = null;

                uint offset = chrvalue.ReadUInt();
                if (offset == 0xFFFFFFFF)
                {
                    SensorCalExpiryValid = false;
                    Debug.WriteLine("Sensor needs to be calibrated - no previous calibration issued, maybe still in warmup");
                }
                else
                {
                    if (!SensorCalExpiryValid)
                    {
                        // this is the first calibration expiry value received
                        if (Listener != null)
                            Listener.OnStatusChange(AppSensorStatus.Warmup);
                    }

                    SensorCalExpiryValid = true;
                    sensorCalExpire = new DateTime(1970, 1, 1, 0, 0, 0, 0, System.DateTimeKind.Utc);
                    sensorCalExpire = sensorCalExpire.Value.AddSeconds(SensorTimeStart + offset).ToLocalTime();

                    Debug.WriteLine("Calibration will expire at date/time: " + sensorCalExpire.Value.ToLongDateString() + ", " + sensorCalExpire.Value.ToLongTimeString());
                }

                if (Listener != null)
                    Listener.OnCalibrationExpireDateChanged(sensorCalExpire);
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Exception occurred (6): " + ex.Message);
                Crashes.TrackError(ex);
            }
        }

        private void OnBatteryLevelValueIndicated(object o, CharacteristicUpdatedEventArgs args)
        {
            var value = args.Characteristic.Value;

            Debug.WriteLine("< BATTERY_CHARGE data: " + Utils.ByteArrayToHexString(value));
            MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "BATTERY_CHARGE", DataString = Utils.ByteArrayToHexString(value) }, SensorCommRecordMessage.Name);

            if (Listener != null)
                Listener.OnBatteryStateChange((double)value[0]);
        }

        private void OnNonCritOpsIndicated(object o, CharacteristicUpdatedEventArgs args)
        {
            var value = args.Characteristic.Value;

            Debug.WriteLine("< NONCRIT_OPS data: " + Utils.ByteArrayToHexString(value));
            MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "NONCRIT_OPS", DataString = Utils.ByteArrayToHexString(value) }, SensorCommRecordMessage.Name);

            // TODO: process data, if there's anything at all to be processed (verify)
        }

        private async void OnCritOpsIndicated(object o, CharacteristicUpdatedEventArgs args)
        {
            try
            {
                var value = args.Characteristic.Value;

                Debug.WriteLine("< CRIT_OPS data: " + Utils.ByteArrayToHexString(value));
                MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "CRIT_OPS", DataString = Utils.ByteArrayToHexString(value) }, SensorCommRecordMessage.Name);

                Opcodes opcode = (Opcodes)value[0];

                if (opcode == Opcodes.AUTH_CHALLENGE && (lastSentCritOpsPacketBytes != null || lastSentNonCritOpsPacketBytes != null))
                {
                    Array.Copy(value, 1, sensorAdvKey, 0, 16);

                    var toSign = lastSentCritOpsPacketBytes;

                    Opcodes lastOpcode = (Opcodes)toSign[0];

                    byte[] usedKey;
                    if (lastOpcode == Opcodes.ENABLE_APP_SECURITY)
                    {
                        usedKey = PresharedGuardianKey;
                        // we assume pairing was successful at this point (althought it could not be true)
                        sensorPaired = true;
                    }
                    else if ((lastOpcode == Opcodes.SYNC_TIME || lastOpcode == Opcodes.REQUEST_CAL_EXPIRATION || lastOpcode == Opcodes.SET_SENSOR_TYPE))
                        usedKey = CryptoApp.GetKey();
                    else
                    {
                        Debug.WriteLine("I don't know how to sign packet with opcode " + lastOpcode);
                        return;
                    }

                    await Send_CritOps(Packets.AuthChallengeResponse.Build(usedKey, toSign, sensorAdvKey), false);

                    lastSentCritOpsPacketBytes = null;
                }
                else if (opcode == Opcodes.SYNC_TIME_RESPONSE)
                {
                    ChrValueWrapper val = new ChrValueWrapper(value);
                    val.ReadByte(); // discard opcode

                    uint timeOffset = val.ReadUInt();
                    byte unkByte = val.ReadByte();

                    SensorTimeStart = DateTime.UtcNow.Subtract(new DateTime(1970, 1, 1)).TotalSeconds - (double)timeOffset;
                    StorageMgr.SetSensorStartTime(SensorTimeStart);

                    DateTime sensorDatetimeStart = new DateTime(1970, 1, 1, 0, 0, 0, 0, System.DateTimeKind.Utc);
                    sensorDatetimeStart = sensorDatetimeStart.AddSeconds(SensorTimeStart).ToLocalTime();

                    Debug.WriteLine("Time synchronized; sensor start date/time: " + sensorDatetimeStart.ToLongDateString() + ", " + sensorDatetimeStart.ToLongTimeString());
                    timeSynced = true;

                    // sync time response is an indicator of successful pairing
                    sensorPaired = true;

                    // start periodic updater
                    PeriodicUpdate(null);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Exception occurred (7): " + ex.Message);
                Crashes.TrackError(ex);
            }
        }

        private async void OnStartCommsValueIndicated(object o, CharacteristicUpdatedEventArgs args)
        {
            try
            {
                Debug.WriteLine("< START_COMMS data: " + Utils.ByteArrayToHexString(args.Characteristic.Value));
                MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = true, ChannelName = "START_COMMS", DataString = Utils.ByteArrayToHexString(args.Characteristic.Value) }, SensorCommRecordMessage.Name);

                ChrValueWrapper val = new ChrValueWrapper(args.Characteristic.Value);

                // e.g. 00000000 00000000 07 01 01 00
                uint compatVersion = val.ReadUInt();
                uint compatExtension = val.ReadUInt();
                byte authLevel = val.ReadByte();
                bool repairRequired = val.ReadByte() != 0;
                bool appSecurityEnabled = val.ReadByte() != 0;
                byte sensorCapabs = val.ReadByte();

                // sensor does not think pairing is necessary, our app does not think either
                if (!repairRequired && sensorPaired)
                {
                    Debug.WriteLine("StartComms: repair is not needed");
                    // apparently everything's OK when the sensor reports "repairRequired == false"
                    if (forceTimeSync)
                    {
                        Debug.WriteLine("StartComms: time sync forced, synchronizing...");
                        forceTimeSync = false;
                        timeSynced = false;
                        SyncTime();
                    }
                    else
                    {
                        timeSynced = true;
                        PeriodicUpdate(null);
                    }
                }
                // our app (sensorPaired) or sensor (authLevel == 0) needs a new pairing procedure
                else if (repairRequired || !sensorPaired || authLevel == 0)
                {
                    Debug.WriteLine("StartComms: repair needed - " + ((repairRequired || (authLevel == 0)) ? "sensor requested" : "app requested"));
                    timeSynced = false;

                    await Send_CritOps(Packets.EnableAppSecurity.Build(KeyExchangeMethod.ENCRYPTED, CryptoApp.GetKey(), ((Support.CounterModeCryptoTransform)EncryptorSensor)), true);
                }
                else if (!timeSynced)
                {
                    Debug.WriteLine("StartComms: repair not needed, but time sync will be done");
                    SyncTime();
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Exception occurred (8): " + ex.Message);
                Crashes.TrackError(ex);
            }
        }

        private async void SyncTime()
        {
            await Send_CritOps(Packets.SyncTime.Build(), true);
        }

        private async Task<Boolean> Send_CritOps(byte[] bytes, bool storeForSignature)
        {
            try
            {
                if (storeForSignature)
                {
                    lastSentCritOpsPacketBytes = new byte[bytes.Length];
                    Array.Copy(bytes, lastSentCritOpsPacketBytes, bytes.Length);
                }

                Debug.WriteLine("> CRIT_OPS write: " + Utils.ByteArrayToHexString(bytes));
                MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = false, ChannelName = "CRIT_OPS", DataString = Utils.ByteArrayToHexString(bytes) }, SensorCommRecordMessage.Name);
                LastSendTime = DateTime.UtcNow;

                await MainThread.InvokeOnMainThreadAsync(async () =>
                {
                    try
                    {
                        //await CritOpsChr.WriteAsync(bytes);
                        await QueuedWrite(CritOpsChr, bytes, true);
                    }
                    catch (Exception ex)
                    {
                        Debug.WriteLine("Could not write characteristic " + CritOpsChr.Id + " due to exception: " + ex.Message);
                    }
                });

                return true;
            }
            catch (Exception)
            {
                return false;
            }
        }

        private async Task<Boolean> Send_NonCritOps(byte[] bytes, bool store)
        {
            try
            {
                if (store)
                {
                    lastSentNonCritOpsPacketBytes = new byte[bytes.Length];
                    Array.Copy(bytes, lastSentNonCritOpsPacketBytes, bytes.Length);
                }

                Debug.WriteLine("> NONCRIT_OPS write: " + Utils.ByteArrayToHexString(bytes));
                MessagingCenter.Send(new SensorCommRecordMessage { IsIncoming = false, ChannelName = "NONCRIT_OPS", DataString = Utils.ByteArrayToHexString(bytes) }, SensorCommRecordMessage.Name);
                LastSendTime = DateTime.UtcNow;

                await MainThread.InvokeOnMainThreadAsync(async () =>
                {
                    try
                    {
                        //await NonCritOpsChr.WriteAsync(bytes);
                        await QueuedWrite(NonCritOpsChr, bytes);
                    }
                    catch (Exception ex)
                    {
                        Debug.WriteLine("Could not write characteristic " + NonCritOpsChr.Id + " due to exception: " + ex.Message);
                    }
                });

                return true;
            }
            catch (Exception)
            {
                return false;
            }
        }

        private async Task<byte[]> ReadCharacteristic(ICharacteristic chr)
        {
            for (int i = 0; i < 3; i++)
            {
                try
                {
                    /*
                    byte[] result = await chr.ReadAsync();
                    if (result != null)
                        return result;
                    */
                    var result = await QueuedRead(chr);
                    if (result != null && result.Success)
                        return result.Data;
                }
                catch (Exception ex)
                {
                    Debug.WriteLine("Could not read characteristic " + chr.Id + " due to exception: " + ex.Message);
                }

                System.Threading.Thread.Sleep(500);
            }

            return null;
        }

    }
}
