﻿using scgms_mobile.Misc;
using Plugin.BLE.Abstractions.Contracts;
using Plugin.BLE.Abstractions.EventArgs;
using System;
using System.Collections.Generic;
using System.Text;
using System.Threading.Tasks;

namespace scgms_mobile.Devices.IEEE11073
{
    /// <summary>
    ///   Implementation of IEEE11073-10417 compliant device
    /// </summary>
    class IEEEDevice : Generic.DeviceAdapterBase
    {
        private DateTime SessionStart;

        /// <summary>
        /// IEEE Device constructor
        /// </summary>
        /// <param name="dev">device to be wrapped</param>
        public IEEEDevice(IDevice dev) : base(dev)
        {
        }

        /// <summary>
        /// Initializes adapter, subscribes for updates, etc.
        /// </summary>
        public override async Task<Boolean> Init()
        {
            IService service = await BLEDevice.GetServiceAsync(BLEKnownGuids.IEEEMainServiceGUID);
            ICharacteristic measChr = await service.GetCharacteristicAsync(BLEKnownGuids.IEEE_MeasurementGUID);

            SessionStart = DateTime.Now;

            // subscribe for measured value changes
            measChr.ValueUpdated += OnMeasuredValueUpdate;
            await measChr.StartUpdatesAsync();

            return true;
        }

        /// <summary>
        /// Parses input from CGM Measurement characteristics and updates listener
        /// </summary>
        /// <param name="input">received bytes from CGM measurement characteristics</param>
        private void UpdateMeasuredValue(byte[] input)
        {
            ChrValueWrapper val = new ChrValueWrapper(input);

            int len = val.ReadByte();
            byte flags = val.ReadByte();
            double measurement = val.ReadSFloat() / 18.018; // convert from mg/dl (defined in IEEE 11073) to mmol/l (our architecture)
            uint seq = val.ReadUShort();

            long measTime = (Int64)(SessionStart.AddSeconds(seq * 60).Subtract(new DateTime(1970, 1, 1)).TotalSeconds);

            // update listener
            if (Listener != null)
                Listener.OnNewMeasurement(Generic.DeviceSignalType.Interstitial_Glucose, measurement, measTime, seq, true);
        }

        /// <summary>
        /// Callback from BLE module - called upon receiving a new value from measurement characteristics
        /// </summary>
        /// <param name="o">source object</param>
        /// <param name="args">event arguments</param>
        private void OnMeasuredValueUpdate(object o, CharacteristicUpdatedEventArgs args)
        {
            UpdateMeasuredValue(args.Characteristic.Value);
        }
    }
}
